/*
 * @(#)JarVerifier.java	1.40 09/12/09                                    //IBM-sun_bug_fixes_temp
 *
 * Copyright 2006-2009 Sun Microsystems, Inc. All rights reserved.              //IBM-sun_bug_fixes_temp
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package java.util.jar;

import java.io.*;
import java.net.URL;                                                            //IBM-sun_bug_fixes_temp
import java.util.*;
import java.util.zip.*;
import java.security.*;
import java.security.cert.CertificateException;

import sun.security.util.ManifestDigester;
import sun.security.util.ManifestEntryVerifier;
import sun.security.util.SignatureFileVerifier;
import sun.security.util.Debug;

/**
 *
 * @version 	1.40 09/12/09                                               //IBM-sun_bug_fixes_temp
 * @author	Roland Schemers
 */
class JarVerifier {

    /*                                                                          //IBM-Harmony_security
     * If using Harmony security then force it to initialize prior to Jar Verification //IBM-Harmony_security
     * to prevent JarVerifer from getting confused during jar file loading      //IBM-Harmony_security
     */                                                                         //IBM-Harmony_security
    static {                                                                    //IBM-Harmony_security
	try {                                                                   //IBM-Harmony_security
	    Class.forName("org.apache.harmony.security.fortress.Services", true, null); //IBM-Harmony_security
	} catch(Exception e) {}                                                 //IBM-Harmony_security
    }                                                                           //IBM-Harmony_security
                                                                                //IBM-Harmony_security
    /* Are we debugging ? */
    static final Debug debug = Debug.getInstance("jar");

    /* a table mapping names to code signers, for jar entries that have
       had their actual hashes verified */
    private Hashtable verifiedSigners;

    /* a table mapping names to code signers, for jar entries that have
       passed the .SF/.DSA -> MANIFEST check */
    private Hashtable sigFileSigners;

    /* a hash table to hold .SF bytes */
    private Hashtable sigFileData;

    /** "queue" of pending PKCS7 blocks that we couldn't parse
     *  until we parsed the .SF file */
    private ArrayList pendingBlocks;

    /* cache of CodeSigner objects */
    private ArrayList signerCache;

    /* Are we parsing a block? */
    private boolean parsingBlockOrSF = false;

    /* Are we done parsing META-INF entries? */
    private boolean parsingMeta = true;

    /* Are there are files to verify? */
    private boolean anyToVerify = true;

    /* The output stream to use when keeping track of files we are interested
       in */
    private ByteArrayOutputStream baos;

    /** The ManifestDigester object */
    private ManifestDigester manDig;

    /** the bytes for the manDig object */
    byte manifestRawBytes[] = null;

    /** controls eager signature validation */                                  //IBM-sun_bug_fixes_temp
    boolean eagerValidation;                                                    //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    /** makes code source singleton instances unique to us */                   //IBM-sun_bug_fixes_temp
    private Object csdomain = new Object();                                     //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    /** collect -DIGEST-MANIFEST values for blacklist */                        //IBM-sun_bug_fixes_temp
    private List manifestDigests;                                               //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    public JarVerifier(byte rawBytes[]) {
	manifestRawBytes = rawBytes;
	sigFileSigners = new Hashtable();
	verifiedSigners = new Hashtable();
	sigFileData = new Hashtable(11);
	pendingBlocks = new ArrayList();
	baos = new ByteArrayOutputStream();
        manifestDigests = new ArrayList();                                      //IBM-sun_bug_fixes_temp
    }

    /**
     * This method scans to see which entry we're parsing and
     * keeps various state information depending on what type of
     * file is being parsed.
     */
    public void beginEntry(JarEntry je, ManifestEntryVerifier mev)
	throws IOException
    {
	if (je == null)
	    return;

	if (debug != null) {
	    debug.println("beginEntry "+je.getName());
	}

	String name = je.getName();

	/*
	 * Assumptions:
	 * 1. The manifest should be the first entry in the META-INF directory.
	 * 2. The .SF/.DSA files follow the manifest, before any normal entries
	 * 3. Any of the following will throw a SecurityException:
	 *    a. digest mismatch between a manifest section and
	 *       the SF section.
	 *    b. digest mismatch between the actual jar entry and the manifest
	 */

	if (parsingMeta) {
	    String uname = name.toUpperCase(Locale.ENGLISH);
	    if ((uname.startsWith("META-INF/") ||
		 uname.startsWith("/META-INF/"))) {

		if (je.isDirectory()) {
		    mev.setEntry(null, je);
		    return;
		}

		if (SignatureFileVerifier.isBlockOrSF(uname)) {
		    /* We parse only DSA or RSA PKCS7 blocks. */
		    parsingBlockOrSF = true;
		    baos.reset();
		    mev.setEntry(null, je);
		}
		return;
	    }
	}

	if (parsingMeta) {
	    doneWithMeta();
	}

	if (je.isDirectory()) {
	    mev.setEntry(null, je);
	    return;
	}

	// be liberal in what you accept. If the name starts with ./, remove
	// it as we internally canonicalize it with out the ./.
	if (name.startsWith("./"))
	    name = name.substring(2);

	// be liberal in what you accept. If the name starts with /, remove
	// it as we internally canonicalize it with out the /.
	if (name.startsWith("/"))
	    name = name.substring(1);

	// only set the jev object for entries that have a signature
	if (sigFileSigners.get(name) != null) {
	    mev.setEntry(name, je);
	    return;
	}

	// don't compute the digest for this entry
	mev.setEntry(null, je);

	return;
    }

    /**
     * update a single byte.
     */

    public void update(int b, ManifestEntryVerifier mev)
	throws IOException
    {
	if (b != -1) {
	    if (parsingBlockOrSF) {
		baos.write(b);
	    } else {
		mev.update((byte)b);
	    }
	} else {
	    processEntry(mev);
	}
    }

    /**
     * update an array of bytes.
     */

    public void update(int n, byte[] b, int off, int len,
		       ManifestEntryVerifier mev)
	throws IOException
    {
	if (n != -1) {
	    if (parsingBlockOrSF) {
		baos.write(b, off, n);
	    } else {
		mev.update(b, off, n);
	    }
	} else {
	    processEntry(mev);
	}
    }

    /**
     * called when we reach the end of entry in one of the read() methods.
     */
    private void processEntry(ManifestEntryVerifier mev)
	throws IOException
    {
	if (!parsingBlockOrSF) {
	    JarEntry je = mev.getEntry();
	    if ((je != null) && (je.signers == null)) {
		je.signers = mev.verify(verifiedSigners, sigFileSigners);
		je.certs = mapSignersToCertArray(je.signers);
	    }
	} else {

	    try {
		parsingBlockOrSF = false;

		if (debug != null) {
		    debug.println("processEntry: processing block");
		}

		String uname = mev.getEntry().getName()
                                             .toUpperCase(Locale.ENGLISH);

		if (uname.endsWith(".SF")) {
		    String key = uname.substring(0, uname.length()-3);
		    byte bytes[] = baos.toByteArray();
		    // add to sigFileData in case future blocks need it
		    sigFileData.put(key, bytes);
		    // check pending blocks, we can now process
		    // anyone waiting for this .SF file
		    Iterator it = pendingBlocks.iterator();
		    while (it.hasNext()) {
			SignatureFileVerifier sfv =
			    (SignatureFileVerifier) it.next();
			if (sfv.needSignatureFile(key)) {
			    if (debug != null) {
				debug.println(
				 "processEntry: processing pending block");
			    }

			    sfv.setSignatureFile(bytes);
			    sfv.process(sigFileSigners, manifestDigests);       //IBM-sun_bug_fixes_temp
			}
		    }
		    return;
		}

		// now we are parsing a signature block file

		String key = uname.substring(0, uname.lastIndexOf("."));

		if (signerCache == null)
		    signerCache = new ArrayList();

		if (manDig == null) {
		    synchronized(manifestRawBytes) {
			if (manDig == null) {
			    manDig = new ManifestDigester(manifestRawBytes);
			    manifestRawBytes = null;
			}
		    }
		}

		SignatureFileVerifier sfv =
		  new SignatureFileVerifier(signerCache,
					    manDig, uname, baos.toByteArray());

		if (sfv.needSignatureFileBytes()) {
		    // see if we have already parsed an external .SF file
		    byte[] bytes = (byte[]) sigFileData.get(key);

		    if (bytes == null) {
			// put this block on queue for later processing
			// since we don't have the .SF bytes yet
			// (uname, block);
			if (debug != null) {
			    debug.println("adding pending block");
			}
			pendingBlocks.add(sfv);
			return;
		    } else {
			sfv.setSignatureFile(bytes);
		    }
		}
		sfv.process(sigFileSigners, manifestDigests);                   //IBM-sun_bug_fixes_temp

	    } catch (sun.security.pkcs.ParsingException pe) {
		if (debug != null) debug.println("processEntry caught: "+pe);
		// ignore and treat as unsigned
	    } catch (IOException ioe) {
		if (debug != null) debug.println("processEntry caught: "+ioe);
		// ignore and treat as unsigned
	    } catch (SignatureException se) {
		if (debug != null) debug.println("processEntry caught: "+se);
		// ignore and treat as unsigned
	    } catch (NoSuchAlgorithmException nsae) {
		if (debug != null) debug.println("processEntry caught: "+nsae);
		// ignore and treat as unsigned
	    } catch (CertificateException ce) {
		if (debug != null) debug.println("processEntry caught: "+ce);
		// ignore and treat as unsigned
	    }
	}
    }

    /**
     * Return an array of java.security.cert.Certificate objects for
     * the given file in the jar. 
     * Deprecated.                                                              //IBM-sun_bug_fixes_temp
     */
    public java.security.cert.Certificate[] getCerts(String name)
    {
	return mapSignersToCertArray(getCodeSigners(name));
    }

    public java.security.cert.Certificate[] getCerts(JarFile jar, JarEntry entry) //IBM-sun_bug_fixes_temp
    {                                                                           //IBM-sun_bug_fixes_temp
	return mapSignersToCertArray(getCodeSigners(jar, entry));               //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    /**
     * return an array of CodeSigner objects for
     * the given file in the jar. this array is not cloned.
     */
    public CodeSigner[] getCodeSigners(String name)
    {
	return (CodeSigner[])verifiedSigners.get(name);
    }

    public CodeSigner[] getCodeSigners(JarFile jar, JarEntry entry)             //IBM-sun_bug_fixes_temp
    {                                                                           //IBM-sun_bug_fixes_temp
	String name = entry.getName();                                          //IBM-sun_bug_fixes_temp
	if (eagerValidation && sigFileSigners.get(name) != null) {              //IBM-sun_bug_fixes_temp
	    /*                                                                  //IBM-sun_bug_fixes_temp
	     * Force a read of the entry data to generate the                   //IBM-sun_bug_fixes_temp
	     * verification hash.                                               //IBM-sun_bug_fixes_temp
	     */                                                                 //IBM-sun_bug_fixes_temp
	    try {                                                               //IBM-sun_bug_fixes_temp
	        InputStream s = jar.getInputStream(entry);                      //IBM-sun_bug_fixes_temp
	        byte[] buffer = new byte[1024];                                 //IBM-sun_bug_fixes_temp
	        int n = buffer.length;                                          //IBM-sun_bug_fixes_temp
	        while (n != -1) {                                               //IBM-sun_bug_fixes_temp
	            n = s.read(buffer, 0, buffer.length);                       //IBM-sun_bug_fixes_temp
	        }                                                               //IBM-sun_bug_fixes_temp
	        s.close();                                                      //IBM-sun_bug_fixes_temp
	    } catch (IOException e) {                                           //IBM-sun_bug_fixes_temp
	    }                                                                   //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
	return getCodeSigners(name);                                            //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    /*
     * Convert an array of signers into an array of concatenated certificate 
     * arrays.
     */
    private static java.security.cert.Certificate[] mapSignersToCertArray(
	CodeSigner[] signers) {

	if (signers != null) {
	    ArrayList certChains = new ArrayList();
	    for (int i = 0; i < signers.length; i++) {
		certChains.addAll(
		    signers[i].getSignerCertPath().getCertificates());
	    }

	    // Convert into a Certificate[]
	    return (java.security.cert.Certificate[])
		certChains.toArray(
		    new java.security.cert.Certificate[certChains.size()]);
	}
	return null;
    }

    /**
     * returns true if there no files to verify.
     * should only be called after all the META-INF entries
     * have been processed.
     */
    boolean nothingToVerify()
    {
	return (anyToVerify == false);
    }

    /**
     * called to let us know we have processed all the
     * META-INF entries, and if we re-read one of them, don't
     * re-process it. Also gets rid of any data structures
     * we needed when parsing META-INF entries.
     */
    void doneWithMeta()
    {
	parsingMeta = false;
	anyToVerify = !sigFileSigners.isEmpty();
	baos = null;
	sigFileData = null;
	pendingBlocks = null;
	signerCache = null;
	manDig = null;
    }

    static class VerifierStream extends java.io.InputStream {

	private InputStream is;
	private JarVerifier jv;
	private ManifestEntryVerifier mev;
	private long numLeft;

	VerifierStream(Manifest man,
		       JarEntry je,
		       InputStream is,
		       JarVerifier jv) throws IOException
	{
	    this.is = is;
	    this.jv = jv;
	    this.mev = new ManifestEntryVerifier(man);
	    this.jv.beginEntry(je, mev);
	    this.numLeft = je.getSize();
	    if (this.numLeft == 0)
		this.jv.update(-1, this.mev);
	}

	public int read() throws IOException
	{
	    if (numLeft > 0) {
		int b = is.read();
		jv.update(b, mev);
		numLeft--;
		if (numLeft == 0)
		    jv.update(-1, mev);
		return b;
	    } else {
		return -1;
	    }
	}

	public int read(byte b[], int off, int len) throws IOException {
	    if ((numLeft > 0) && (numLeft < len)) {
		len = (int)numLeft;
	    }

	    if (numLeft > 0) {
		int n = is.read(b, off, len);
		jv.update(n, b, off, len, mev);
		numLeft -= n;
		if (numLeft == 0)
		    jv.update(-1, b, off, len, mev);
		return n;
	    } else {
		return -1;
	    }
	}

	public void close()
	    throws IOException
	{
	    if (is != null)
		is.close();
	    is = null;
	    mev = null;
	    jv = null;
	}

	public int available() throws IOException {
	    return is.available();
	}

    }
                                                                                //IBM-sun_bug_fixes_temp
    // Extended JavaUtilJarAccess CodeSource API Support                        //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    private Map urlToCodeSourceMap = new HashMap();                             //IBM-sun_bug_fixes_temp
    private Map signerToCodeSource = new HashMap();                             //IBM-sun_bug_fixes_temp
    private URL lastURL;                                                        //IBM-sun_bug_fixes_temp
    private Map lastURLMap;                                                     //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    /*                                                                          //IBM-sun_bug_fixes_temp
     * Create a unique mapping from codeSigner cache entries to CodeSource.     //IBM-sun_bug_fixes_temp
     * In theory, multiple URLs origins could map to a single locally cached    //IBM-sun_bug_fixes_temp
     * and shared JAR file although in practice there will be a single URL in use. //IBM-sun_bug_fixes_temp
     */                                                                         //IBM-sun_bug_fixes_temp
    private synchronized CodeSource mapSignersToCodeSource(URL url, CodeSigner[] signers) { //IBM-sun_bug_fixes_temp
	Map map;                                                                //IBM-sun_bug_fixes_temp
	if (url == lastURL) {                                                   //IBM-sun_bug_fixes_temp
	    map = lastURLMap;                                                   //IBM-sun_bug_fixes_temp
	} else {                                                                //IBM-sun_bug_fixes_temp
	    map = (Map)urlToCodeSourceMap.get(url);                             //IBM-sun_bug_fixes_temp
	    if (map == null) {                                                  //IBM-sun_bug_fixes_temp
		map = new HashMap();                                            //IBM-sun_bug_fixes_temp
		urlToCodeSourceMap.put(url, map);                               //IBM-sun_bug_fixes_temp
	    }                                                                   //IBM-sun_bug_fixes_temp
	    lastURLMap = map;                                                   //IBM-sun_bug_fixes_temp
	    lastURL = url;                                                      //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
	CodeSource cs = (CodeSource) map.get(signers);                          //IBM-sun_bug_fixes_temp
	if (cs == null) {                                                       //IBM-sun_bug_fixes_temp
	    cs = new VerifierCodeSource(csdomain, url, signers);                //IBM-sun_bug_fixes_temp
	    signerToCodeSource.put(signers, cs);                                //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
	return cs;                                                              //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    private CodeSource[] mapSignersToCodeSources(URL url, List signers, boolean unsigned) { //IBM-sun_bug_fixes_temp
	List sources = new ArrayList();                                         //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
	for (int i=0; i < signers.size(); i++) {                                //IBM-sun_bug_fixes_temp
	    sources.add(mapSignersToCodeSource(url, (CodeSigner[]) signers.get(i))); //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
	if (unsigned) {                                                         //IBM-sun_bug_fixes_temp
	    sources.add(mapSignersToCodeSource(url, null));                     //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
	return (CodeSource[])sources.toArray(new CodeSource[sources.size()]);   //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    private CodeSigner[] emptySigner = new CodeSigner[0];                       //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    /*                                                                          //IBM-sun_bug_fixes_temp
     * Match CodeSource to a CodeSigner[] in the signer cache.                  //IBM-sun_bug_fixes_temp
     */                                                                         //IBM-sun_bug_fixes_temp
    private CodeSigner[] findMatchingSigners(CodeSource cs) {                   //IBM-sun_bug_fixes_temp
	if (cs instanceof VerifierCodeSource) {                                 //IBM-sun_bug_fixes_temp
	    VerifierCodeSource vcs = (VerifierCodeSource) cs;                   //IBM-sun_bug_fixes_temp
	    if (vcs.isSameDomain(csdomain)) {                                   //IBM-sun_bug_fixes_temp
	        return ((VerifierCodeSource)cs).getPrivateSigners();            //IBM-sun_bug_fixes_temp
	    }                                                                   //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
	/*                                                                      //IBM-sun_bug_fixes_temp
         * In practice signers should always be optimized above                 //IBM-sun_bug_fixes_temp
	 * but this handles a CodeSource of any type, just in case.             //IBM-sun_bug_fixes_temp
	 */                                                                     //IBM-sun_bug_fixes_temp
	CodeSource[] sources = mapSignersToCodeSources(cs.getLocation(), getJarCodeSigners(), true); //IBM-sun_bug_fixes_temp
	List sourceList = new ArrayList();                                      //IBM-sun_bug_fixes_temp
	for (int i=0; i < sources.length; i++) {                                //IBM-sun_bug_fixes_temp
	    sourceList.add(sources[i]);                                         //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
	int j = sourceList.indexOf(cs);                                         //IBM-sun_bug_fixes_temp
	if (j != -1) {                                                          //IBM-sun_bug_fixes_temp
	    CodeSigner[] match;                                                 //IBM-sun_bug_fixes_temp
	    match = ((VerifierCodeSource)sourceList.get(j)).getPrivateSigners(); //IBM-sun_bug_fixes_temp
	    if (match == null) {                                                //IBM-sun_bug_fixes_temp
		match = emptySigner;                                            //IBM-sun_bug_fixes_temp
	    }                                                                   //IBM-sun_bug_fixes_temp
	    return match;                                                       //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
	return null;                                                            //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    /*                                                                          //IBM-sun_bug_fixes_temp
     * Instances of this class hold uncopied references to internal             //IBM-sun_bug_fixes_temp
     * signing data that can be compared by object reference identity.          //IBM-sun_bug_fixes_temp
     */                                                                         //IBM-sun_bug_fixes_temp
    private static class VerifierCodeSource extends CodeSource {                //IBM-sun_bug_fixes_temp
	URL vlocation;                                                          //IBM-sun_bug_fixes_temp
	CodeSigner[] vsigners;                                                  //IBM-sun_bug_fixes_temp
	java.security.cert.Certificate[] vcerts;                                //IBM-sun_bug_fixes_temp
	Object csdomain;                                                        //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
	VerifierCodeSource(Object csdomain, URL location, CodeSigner[] signers) { //IBM-sun_bug_fixes_temp
	    super(location, signers);                                           //IBM-sun_bug_fixes_temp
	    this.csdomain = csdomain;                                           //IBM-sun_bug_fixes_temp
	    vlocation = location;                                               //IBM-sun_bug_fixes_temp
	    vsigners = signers; // from signerCache                             //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
	VerifierCodeSource(Object csdomain, URL location, java.security.cert.Certificate[] certs) { //IBM-sun_bug_fixes_temp
	    super(location, certs);                                             //IBM-sun_bug_fixes_temp
	    this.csdomain = csdomain;                                           //IBM-sun_bug_fixes_temp
	    vlocation = location;                                               //IBM-sun_bug_fixes_temp
	    vcerts = certs; // from signerCache                                 //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
	/*                                                                      //IBM-sun_bug_fixes_temp
	 * All VerifierCodeSource instances are constructed based on            //IBM-sun_bug_fixes_temp
	 * singleton signerCache or signerCacheCert entries for each unique signer. //IBM-sun_bug_fixes_temp
	 * No CodeSigner<->Certificate[] conversion is required.                //IBM-sun_bug_fixes_temp
	 * We use these assumptions to optimize equality comparisons.           //IBM-sun_bug_fixes_temp
	 */                                                                     //IBM-sun_bug_fixes_temp
	public boolean equals(Object obj) {                                     //IBM-sun_bug_fixes_temp
	    if (obj == this) {                                                  //IBM-sun_bug_fixes_temp
		return true;                                                    //IBM-sun_bug_fixes_temp
	    }                                                                   //IBM-sun_bug_fixes_temp
	    if (obj instanceof VerifierCodeSource) {                            //IBM-sun_bug_fixes_temp
		VerifierCodeSource that = (VerifierCodeSource) obj;             //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
		/*                                                              //IBM-sun_bug_fixes_temp
		 * Only compare against other per-signer singletons constructed //IBM-sun_bug_fixes_temp
		 * on behalf of the same JarFile instance. Otherwise, compare   //IBM-sun_bug_fixes_temp
		 * things the slower way.                                       //IBM-sun_bug_fixes_temp
		 */                                                             //IBM-sun_bug_fixes_temp
		if (isSameDomain(that.csdomain)) {                              //IBM-sun_bug_fixes_temp
		    if (that.vsigners != this.vsigners ||                       //IBM-sun_bug_fixes_temp
		        that.vcerts != this.vcerts) {                           //IBM-sun_bug_fixes_temp
		        return false;                                           //IBM-sun_bug_fixes_temp
		    }                                                           //IBM-sun_bug_fixes_temp
		    if (that.vlocation != null) {                               //IBM-sun_bug_fixes_temp
		        return that.vlocation.equals(this.vlocation);           //IBM-sun_bug_fixes_temp
		    } else if (this.vlocation != null) {                        //IBM-sun_bug_fixes_temp
		        return this.vlocation.equals(that.vlocation);           //IBM-sun_bug_fixes_temp
		    } else { // both null                                       //IBM-sun_bug_fixes_temp
		        return true;                                            //IBM-sun_bug_fixes_temp
		    }                                                           //IBM-sun_bug_fixes_temp
		}                                                               //IBM-sun_bug_fixes_temp
	    }                                                                   //IBM-sun_bug_fixes_temp
	    return super.equals(obj);                                           //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
	boolean isSameDomain(Object csdomain) {                                 //IBM-sun_bug_fixes_temp
	    return this.csdomain == csdomain;                                   //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
	private CodeSigner[] getPrivateSigners() {                              //IBM-sun_bug_fixes_temp
	    return vsigners;                                                    //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
	private java.security.cert.Certificate[] getPrivateCertificates() {     //IBM-sun_bug_fixes_temp
	    return vcerts;                                                      //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    private Map signerMap;                                                      //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    private synchronized Map signerMap() {                                      //IBM-sun_bug_fixes_temp
	if (signerMap == null) {                                                //IBM-sun_bug_fixes_temp
	    /*                                                                  //IBM-sun_bug_fixes_temp
	     * Snapshot signer state so it doesn't change on us. We care        //IBM-sun_bug_fixes_temp
	     * only about the asserted signatures. Verification of              //IBM-sun_bug_fixes_temp
	     * signature validity happens via the JarEntry apis.                //IBM-sun_bug_fixes_temp
	     */                                                                 //IBM-sun_bug_fixes_temp
	    signerMap = new HashMap(verifiedSigners.size() + sigFileSigners.size()); //IBM-sun_bug_fixes_temp
	    signerMap.putAll(verifiedSigners);                                  //IBM-sun_bug_fixes_temp
	    signerMap.putAll(sigFileSigners);                                   //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
	return signerMap;                                                       //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    public synchronized Enumeration<String> entryNames(JarFile jar, final CodeSource[] cs) { //IBM-sun_bug_fixes_temp
	final Map map = signerMap();                                            //IBM-sun_bug_fixes_temp
	final Iterator itor = map.entrySet().iterator();                        //IBM-sun_bug_fixes_temp
	boolean matchUnsigned = false;                                          //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
        /*                                                                      //IBM-sun_bug_fixes_temp
         * Grab a single copy of the CodeSigner arrays. Check                   //IBM-sun_bug_fixes_temp
         * to see if we can optimize CodeSigner equality test.                  //IBM-sun_bug_fixes_temp
         */                                                                     //IBM-sun_bug_fixes_temp
        List req = new ArrayList(cs.length);                                    //IBM-sun_bug_fixes_temp
        for (int i=0; i < cs.length; i++) {                                     //IBM-sun_bug_fixes_temp
	    CodeSigner[] match = findMatchingSigners(cs[i]);                    //IBM-sun_bug_fixes_temp
	    if (match != null) {                                                //IBM-sun_bug_fixes_temp
		if (match.length > 0) {                                         //IBM-sun_bug_fixes_temp
                    req.add(match);                                             //IBM-sun_bug_fixes_temp
		} else {                                                        //IBM-sun_bug_fixes_temp
		    matchUnsigned = true;                                       //IBM-sun_bug_fixes_temp
		}                                                               //IBM-sun_bug_fixes_temp
	    }                                                                   //IBM-sun_bug_fixes_temp
        }                                                                       //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
	final List signersReq = req;                                            //IBM-sun_bug_fixes_temp
	final Enumeration enum2 = (matchUnsigned) ? unsignedEntryNames(jar) : emptyEnumeration; //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
	return new Enumeration<String>() {                                      //IBM-sun_bug_fixes_temp
	    String name;                                                        //IBM-sun_bug_fixes_temp
	                                                                        //IBM-sun_bug_fixes_temp
	    public boolean hasMoreElements() {                                  //IBM-sun_bug_fixes_temp
		if (name != null) {                                             //IBM-sun_bug_fixes_temp
		    return true;                                                //IBM-sun_bug_fixes_temp
		}                                                               //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
		while (itor.hasNext()) {                                        //IBM-sun_bug_fixes_temp
		    Map.Entry e = (Map.Entry)itor.next();                       //IBM-sun_bug_fixes_temp
		    if (signersReq.contains((CodeSigner[])e.getValue())) {      //IBM-sun_bug_fixes_temp
			name = (String)e.getKey();                              //IBM-sun_bug_fixes_temp
			return true;                                            //IBM-sun_bug_fixes_temp
		    }                                                           //IBM-sun_bug_fixes_temp
		}                                                               //IBM-sun_bug_fixes_temp
		while (enum2.hasMoreElements()) {                               //IBM-sun_bug_fixes_temp
		    name = (String)enum2.nextElement();                         //IBM-sun_bug_fixes_temp
		    return true;                                                //IBM-sun_bug_fixes_temp
		}                                                               //IBM-sun_bug_fixes_temp
		return false;                                                   //IBM-sun_bug_fixes_temp
	    }                                                                   //IBM-sun_bug_fixes_temp
	    public String nextElement() {                                       //IBM-sun_bug_fixes_temp
		if (hasMoreElements()) {                                        //IBM-sun_bug_fixes_temp
		    String value = name;                                        //IBM-sun_bug_fixes_temp
		    name = null;                                                //IBM-sun_bug_fixes_temp
		    return value;                                               //IBM-sun_bug_fixes_temp
		}                                                               //IBM-sun_bug_fixes_temp
		throw new NoSuchElementException();                             //IBM-sun_bug_fixes_temp
	    }                                                                   //IBM-sun_bug_fixes_temp
	};                                                                      //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    /*                                                                          //IBM-sun_bug_fixes_temp
     * Like entries() but screens out internal JAR mechanism entries            //IBM-sun_bug_fixes_temp
     * and includes signed entries with no ZIP data.                            //IBM-sun_bug_fixes_temp
     */                                                                         //IBM-sun_bug_fixes_temp
    public Enumeration<JarEntry> entries2(final JarFile jar, Enumeration e) {   //IBM-sun_bug_fixes_temp
	final Map map = new HashMap();                                          //IBM-sun_bug_fixes_temp
	map.putAll(signerMap());                                                //IBM-sun_bug_fixes_temp
	final Enumeration enum_ = e;                                            //IBM-sun_bug_fixes_temp
	return new Enumeration<JarEntry>() {                                    //IBM-sun_bug_fixes_temp
	    Enumeration signers = null;                                         //IBM-sun_bug_fixes_temp
	    JarEntry entry;                                                     //IBM-sun_bug_fixes_temp
	    public boolean hasMoreElements() {                                  //IBM-sun_bug_fixes_temp
		if (entry != null) {                                            //IBM-sun_bug_fixes_temp
		    return true;                                                //IBM-sun_bug_fixes_temp
		}                                                               //IBM-sun_bug_fixes_temp
		while (enum_.hasMoreElements()) {                               //IBM-sun_bug_fixes_temp
		    ZipEntry ze = (ZipEntry) enum_.nextElement();               //IBM-sun_bug_fixes_temp
		    if (JarVerifier.isSigningRelated(ze.getName())) {           //IBM-sun_bug_fixes_temp
			continue;                                               //IBM-sun_bug_fixes_temp
		    }                                                           //IBM-sun_bug_fixes_temp
		    entry = jar.newEntry(ze);                                   //IBM-sun_bug_fixes_temp
		    return true;                                                //IBM-sun_bug_fixes_temp
		}                                                               //IBM-sun_bug_fixes_temp
		if (signers == null) {                                          //IBM-sun_bug_fixes_temp
		    signers = Collections.enumeration(map.keySet());            //IBM-sun_bug_fixes_temp
		}                                                               //IBM-sun_bug_fixes_temp
		while (signers.hasMoreElements()) {                             //IBM-sun_bug_fixes_temp
		    String name = (String)signers.nextElement();                //IBM-sun_bug_fixes_temp
		    entry = jar.newEntry(new ZipEntry(name));                   //IBM-sun_bug_fixes_temp
		    return true;                                                //IBM-sun_bug_fixes_temp
		}                                                               //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
		// Any map entries left?                                        //IBM-sun_bug_fixes_temp
		return false;                                                   //IBM-sun_bug_fixes_temp
	    }                                                                   //IBM-sun_bug_fixes_temp
	    public JarEntry nextElement() {                                     //IBM-sun_bug_fixes_temp
                if (hasMoreElements()) {                                        //IBM-sun_bug_fixes_temp
                     JarEntry je = entry;                                       //IBM-sun_bug_fixes_temp
		     map.remove(je.getName());                                  //IBM-sun_bug_fixes_temp
                     entry = null;                                              //IBM-sun_bug_fixes_temp
                     return je;                                                 //IBM-sun_bug_fixes_temp
                 }                                                              //IBM-sun_bug_fixes_temp
                 throw new NoSuchElementException();                            //IBM-sun_bug_fixes_temp
	    }                                                                   //IBM-sun_bug_fixes_temp
	};                                                                      //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    private Enumeration emptyEnumeration = new Enumeration<String>() {          //IBM-sun_bug_fixes_temp
        public boolean hasMoreElements() {                                      //IBM-sun_bug_fixes_temp
	    return false;                                                       //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
	public String nextElement() {                                           //IBM-sun_bug_fixes_temp
	    throw new NoSuchElementException();                                 //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
    };                                                                          //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    // true if file is part of the signature mechanism itself                   //IBM-sun_bug_fixes_temp
    static boolean isSigningRelated(String name) {                              //IBM-sun_bug_fixes_temp
        name = name.toUpperCase(Locale.ENGLISH);                                //IBM-sun_bug_fixes_temp
	if (!name.startsWith("META-INF/")) {                                    //IBM-sun_bug_fixes_temp
	    return false;                                                       //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
	name = name.substring(9);                                               //IBM-sun_bug_fixes_temp
	if (name.indexOf('/') != -1) {                                          //IBM-sun_bug_fixes_temp
	    return false;                                                       //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
        if (name.endsWith(".DSA") ||                                            //IBM-sun_bug_fixes_temp
            name.endsWith(".RSA") ||                                            //IBM-sun_bug_fixes_temp
            name.endsWith(".SF")  ||                                            //IBM-sun_bug_fixes_temp
	    name.startsWith("SIG-") ||                                          //IBM-sun_bug_fixes_temp
	    name.equals("MANIFEST.MF")) {                                       //IBM-sun_bug_fixes_temp
	    return true;                                                        //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
	return false;                                                           //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    private Enumeration<String> unsignedEntryNames(JarFile jar) {               //IBM-sun_bug_fixes_temp
	final Map map = signerMap();                                            //IBM-sun_bug_fixes_temp
	final Enumeration entries = jar.entries();                              //IBM-sun_bug_fixes_temp
	return new Enumeration<String>() {                                      //IBM-sun_bug_fixes_temp
	    String name;                                                        //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
	    /*                                                                  //IBM-sun_bug_fixes_temp
	     * Grab entries from ZIP directory but screen out                   //IBM-sun_bug_fixes_temp
	     * metadata.                                                        //IBM-sun_bug_fixes_temp
	     */                                                                 //IBM-sun_bug_fixes_temp
	    public boolean hasMoreElements() {                                  //IBM-sun_bug_fixes_temp
		if (name != null) {                                             //IBM-sun_bug_fixes_temp
		    return true;                                                //IBM-sun_bug_fixes_temp
		}                                                               //IBM-sun_bug_fixes_temp
		while (entries.hasMoreElements()) {                             //IBM-sun_bug_fixes_temp
		    String value;                                               //IBM-sun_bug_fixes_temp
		    ZipEntry e = (ZipEntry) entries.nextElement();              //IBM-sun_bug_fixes_temp
		    value = e.getName();                                        //IBM-sun_bug_fixes_temp
		    if (e.isDirectory() || isSigningRelated(value)) {           //IBM-sun_bug_fixes_temp
			continue;                                               //IBM-sun_bug_fixes_temp
		    }                                                           //IBM-sun_bug_fixes_temp
	    	    if (map.get(value) == null) {                           //IBM-sun_bug_fixes_temp
			name = value;                                           //IBM-sun_bug_fixes_temp
			return true;                                            //IBM-sun_bug_fixes_temp
		    }                                                           //IBM-sun_bug_fixes_temp
		}                                                               //IBM-sun_bug_fixes_temp
		return false;                                                   //IBM-sun_bug_fixes_temp
	    }                                                                   //IBM-sun_bug_fixes_temp
	    public String nextElement() {                                       //IBM-sun_bug_fixes_temp
                if (hasMoreElements()) {                                        //IBM-sun_bug_fixes_temp
                     String value = name;                                       //IBM-sun_bug_fixes_temp
                     name = null;                                               //IBM-sun_bug_fixes_temp
                     return value;                                              //IBM-sun_bug_fixes_temp
                 }                                                              //IBM-sun_bug_fixes_temp
                 throw new NoSuchElementException();                            //IBM-sun_bug_fixes_temp
	    }                                                                   //IBM-sun_bug_fixes_temp
	};                                                                      //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    private List jarCodeSigners;                                                //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    private synchronized List getJarCodeSigners() {                             //IBM-sun_bug_fixes_temp
	CodeSigner[] signers;                                                   //IBM-sun_bug_fixes_temp
	if (jarCodeSigners == null) {                                           //IBM-sun_bug_fixes_temp
	    HashSet set = new HashSet();                                        //IBM-sun_bug_fixes_temp
	    set.addAll(signerMap().values());                                   //IBM-sun_bug_fixes_temp
	    jarCodeSigners = new ArrayList();                                   //IBM-sun_bug_fixes_temp
	    jarCodeSigners.addAll(set);                                         //IBM-sun_bug_fixes_temp
	}                                                                       //IBM-sun_bug_fixes_temp
	return jarCodeSigners;                                                  //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    public synchronized CodeSource[] getCodeSources(JarFile jar, URL url) {     //IBM-sun_bug_fixes_temp
	boolean hasUnsigned = unsignedEntryNames(jar).hasMoreElements();        //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
	return mapSignersToCodeSources(url, getJarCodeSigners(), hasUnsigned);  //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    public CodeSource getCodeSource(URL url, String name) {                     //IBM-sun_bug_fixes_temp
	CodeSigner[] signers;                                                   //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
	signers = (CodeSigner[]) signerMap().get(name);                         //IBM-sun_bug_fixes_temp
	return mapSignersToCodeSource(url, signers);                            //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    public CodeSource getCodeSource(URL url, JarFile jar, JarEntry je) {        //IBM-sun_bug_fixes_temp
	CodeSigner[] signers;                                                   //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
	return mapSignersToCodeSource(url, getCodeSigners(jar, je));            //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    public void setEagerValidation(boolean eager) {                             //IBM-sun_bug_fixes_temp
	eagerValidation = eager;                                                //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    public synchronized List getManifestDigests() {                             //IBM-sun_bug_fixes_temp
	return Collections.unmodifiableList(manifestDigests);                   //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
                                                                                //IBM-sun_bug_fixes_temp
    static CodeSource getUnsignedCS(URL url) {                                  //IBM-sun_bug_fixes_temp
	return new VerifierCodeSource(null, url, (java.security.cert.Certificate[]) null); //IBM-sun_bug_fixes_temp
    }                                                                           //IBM-sun_bug_fixes_temp
}
//IBM-Harmony_security
//IBM-sun_bug_fixes_temp
