/*
 * @(#)Deflater.java	1.45 06/04/07
 *
 * Copyright 2006 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package java.util.zip;

/**
 * This class provides support for general purpose compression using the
 * popular ZLIB compression library. The ZLIB compression library was
 * initially developed as part of the PNG graphics standard and is not
 * protected by patents. It is fully described in the specifications at 
 * the <a href="package-summary.html#package_description">java.util.zip
 * package description</a>.
 *
 * <p>The following code fragment demonstrates a trivial compression
 * and decompression of a string using <tt>Deflater</tt> and
 * <tt>Inflater</tt>.
 *
 * <blockquote><pre>
 * try {
 *     // Encode a String into bytes
 *     String inputString = "blahblahblah\u20AC\u20AC";
 *     byte[] input = inputString.getBytes("UTF-8");
 *
 *     // Compress the bytes
 *     byte[] output = new byte[100];
 *     Deflater compresser = new Deflater();
 *     compresser.setInput(input);
 *     compresser.finish();
 *     int compressedDataLength = compresser.deflate(output);
 *
 *     // Decompress the bytes
 *     Inflater decompresser = new Inflater();
 *     decompresser.setInput(output, 0, compressedDataLength);
 *     byte[] result = new byte[100];
 *     int resultLength = decompresser.inflate(result);
 *     decompresser.end();
 *
 *     // Decode the bytes into a String
 *     String outputString = new String(result, 0, resultLength, "UTF-8");
 * } catch(java.io.UnsupportedEncodingException ex) {
 *     // handle
 * } catch (java.util.zip.DataFormatException ex) {
 *     // handle
 * }
 * </pre></blockquote>
 *
 * @see		Inflater
 * @version 	1.45, 04/07/06
 * @author 	David Connelly
 */
public
class Deflater {
    private final ZStreamRef zsRef;                                             //IBM-security_bulletin_tmp
    private byte[] buf = new byte[0];
    private int off, len;
    private int level, strategy;
    private boolean setParams;
    private boolean finish, finished;

    /**
     * Compression method for the deflate algorithm (the only one currently
     * supported).
     */
    public static final int DEFLATED = 8;

    /**
     * Compression level for no compression.
     */
    public static final int NO_COMPRESSION = 0;

    /**
     * Compression level for fastest compression.
     */
    public static final int BEST_SPEED = 1;

    /**
     * Compression level for best compression.
     */
    public static final int BEST_COMPRESSION = 9;

    /**
     * Default compression level.
     */
    public static final int DEFAULT_COMPRESSION = -1;

    /**
     * Compression strategy best used for data consisting mostly of small
     * values with a somewhat random distribution. Forces more Huffman coding
     * and less string matching.
     */
    public static final int FILTERED = 1;

    /**
     * Compression strategy for Huffman coding only.
     */
    public static final int HUFFMAN_ONLY = 2;

    /**
     * Default compression strategy.
     */
    public static final int DEFAULT_STRATEGY = 0;

    static {
	/* Zip library is loaded from System.initializeSystemClass */
	initIDs();
    }

    /**
     * Creates a new compressor using the specified compression level.
     * If 'nowrap' is true then the ZLIB header and checksum fields will
     * not be used in order to support the compression format used in
     * both GZIP and PKZIP.
     * @param level the compression level (0-9)
     * @param nowrap if true then use GZIP compatible compression
     */
    public Deflater(int level, boolean nowrap) {
	this.level = level;
	this.strategy = DEFAULT_STRATEGY;
	this.zsRef = new ZStreamRef(init(level, DEFAULT_STRATEGY, nowrap));     //IBM-security_bulletin_tmp
    }

    /** 
     * Creates a new compressor using the specified compression level.
     * Compressed data will be generated in ZLIB format.
     * @param level the compression level (0-9)
     */
    public Deflater(int level) {
	this(level, false);
    }

    /**
     * Creates a new compressor with the default compression level.
     * Compressed data will be generated in ZLIB format.
     */
    public Deflater() {
	this(DEFAULT_COMPRESSION, false);
    }

    /**
     * Sets input data for compression. This should be called whenever
     * needsInput() returns true indicating that more input data is required.
     * @param b the input data bytes
     * @param off the start offset of the data
     * @param len the length of the data
     * @see Deflater#needsInput
     */
    public void setInput(byte[] b, int off, int len) {                          //IBM-security_bulletin_tmp
	if (b== null) {
	    throw new NullPointerException();
	}
        if (off < 0 || len < 0 || off > b.length - len) {
	    throw new ArrayIndexOutOfBoundsException();
	}
        synchronized (zsRef) {                                                  //IBM-security_bulletin_tmp
	    this.buf = b;                                                       //IBM-security_bulletin_tmp
	    this.off = off;                                                     //IBM-security_bulletin_tmp
	    this.len = len;                                                     //IBM-security_bulletin_tmp
        }                                                                       //IBM-security_bulletin_tmp
    }

    /**
     * Sets input data for compression. This should be called whenever
     * needsInput() returns true indicating that more input data is required.
     * @param b the input data bytes
     * @see Deflater#needsInput
     */
    public void setInput(byte[] b) {
	setInput(b, 0, b.length);
    }

    /**
     * Sets preset dictionary for compression. A preset dictionary is used
     * when the history buffer can be predetermined. When the data is later
     * uncompressed with Inflater.inflate(), Inflater.getAdler() can be called
     * in order to get the Adler-32 value of the dictionary required for
     * decompression.
     * @param b the dictionary data bytes
     * @param off the start offset of the data
     * @param len the length of the data
     * @see Inflater#inflate
     * @see Inflater#getAdler
     */
    public void setDictionary(byte[] b, int off, int len) {                     //IBM-security_bulletin_tmp
	if (b == null) {                                                        //IBM-security_bulletin_tmp
	    throw new NullPointerException();
	}
        if (off < 0 || len < 0 || off > b.length - len) {
	    throw new ArrayIndexOutOfBoundsException();
	}
        synchronized (zsRef){                                                   //IBM-security_bulletin_tmp
            ensureOpen();                                                       //IBM-security_bulletin_tmp
	    setDictionary(zsRef.address(), b, off, len);                        //IBM-security_bulletin_tmp
        }                                                                       //IBM-security_bulletin_tmp
                                                                                //IBM-security_bulletin_tmp
    }

    /**
     * Sets preset dictionary for compression. A preset dictionary is used
     * when the history buffer can be predetermined. When the data is later
     * uncompressed with Inflater.inflate(), Inflater.getAdler() can be called
     * in order to get the Adler-32 value of the dictionary required for
     * decompression.
     * @param b the dictionary data bytes
     * @see Inflater#inflate
     * @see Inflater#getAdler
     */
    public void setDictionary(byte[] b) {
	setDictionary(b, 0, b.length);
    }

    /**
     * Sets the compression strategy to the specified value.
     * @param strategy the new compression strategy
     * @exception IllegalArgumentException if the compression strategy is
     *				           invalid
     */
    public void setStrategy(int strategy) {                                     //IBM-security_bulletin_tmp
	switch (strategy) {
	  case DEFAULT_STRATEGY:
	  case FILTERED:
	  case HUFFMAN_ONLY:
	    break;
	  default:
	    throw new IllegalArgumentException();
	}
        synchronized (zsRef){                                                   //IBM-security_bulletin_tmp
	    if (this.strategy != strategy) {                                    //IBM-security_bulletin_tmp
	       this.strategy = strategy;                                        //IBM-security_bulletin_tmp
	       setParams = true;                                                //IBM-security_bulletin_tmp
	    }                                                                   //IBM-security_bulletin_tmp
        }                                                                       //IBM-security_bulletin_tmp
    }

    /**
     * Sets the current compression level to the specified value.
     * @param level the new compression level (0-9)
     * @exception IllegalArgumentException if the compression level is invalid
     */
    public void setLevel(int level) {                                           //IBM-security_bulletin_tmp
	if ((level < 0 || level > 9) && level != DEFAULT_COMPRESSION) {
	    throw new IllegalArgumentException("invalid compression level");
	}
        synchronized (zsRef){                                                   //IBM-security_bulletin_tmp
	   if (this.level != level) {                                           //IBM-security_bulletin_tmp
	    this.level = level;
	    setParams = true;
	   }                                                                    //IBM-security_bulletin_tmp
        }                                                                       //IBM-security_bulletin_tmp
    }

    /**
     * Returns true if the input data buffer is empty and setInput()
     * should be called in order to provide more input.
     * @return true if the input data buffer is empty and setInput()
     * should be called in order to provide more input
     */
    public boolean needsInput() {
	return len <= 0;
    }

    /**
     * When called, indicates that compression should end with the current
     * contents of the input buffer.
     */
    public void finish() {                                                      //IBM-security_bulletin_tmp
        synchronized (zsRef) {                                                  //IBM-security_bulletin_tmp
	    finish = true;                                                      //IBM-security_bulletin_tmp
        }                                                                       //IBM-security_bulletin_tmp
    }

    /**
     * Returns true if the end of the compressed data output stream has
     * been reached.
     * @return true if the end of the compressed data output stream has
     * been reached
     */
    public boolean finished() {                                                 //IBM-security_bulletin_tmp
        synchronized (zsRef) {                                                  //IBM-security_bulletin_tmp
	    return finished;                                                    //IBM-security_bulletin_tmp
        }                                                                       //IBM-security_bulletin_tmp
    }

    /**
     * Fills specified buffer with compressed data. Returns actual number
     * of bytes of compressed data. A return value of 0 indicates that
     * needsInput() should be called in order to determine if more input
     * data is required.
     * @param b the buffer for the compressed data
     * @param off the start offset of the data
     * @param len the maximum number of bytes of compressed data
     * @return the actual number of bytes of compressed data
     */
    public int deflate(byte[] b, int off, int len) {                            //IBM-security_bulletin_tmp
	if (b == null) {
	    throw new NullPointerException();
	}
        if (off < 0 || len < 0 || off > b.length - len) {
	    throw new ArrayIndexOutOfBoundsException();
	}
        synchronized (zsRef) {                                                  //IBM-security_bulletin_tmp
            ensureOpen();                                                       //IBM-security_bulletin_tmp
	    return deflateBytes(zsRef.address(),b,off, len);                    //IBM-security_bulletin_tmp
        }                                                                       //IBM-security_bulletin_tmp
    }

    /**
     * Fills specified buffer with compressed data. Returns actual number
     * of bytes of compressed data. A return value of 0 indicates that
     * needsInput() should be called in order to determine if more input
     * data is required.
     * @param b the buffer for the compressed data
     * @return the actual number of bytes of compressed data
     */
    public int deflate(byte[] b) {
	return deflate(b, 0, b.length);
    }

    /**
     * Returns the ADLER-32 value of the uncompressed data.
     * @return the ADLER-32 value of the uncompressed data
     */
    public int getAdler() {                                                     //IBM-security_bulletin_tmp
        synchronized (zsRef) {                                                  //IBM-security_bulletin_tmp
	    ensureOpen();                                                       //IBM-security_bulletin_tmp
	    return getAdler(zsRef.address());                                   //IBM-security_bulletin_tmp
        }                                                                       //IBM-security_bulletin_tmp
    }

    /**
     * Returns the total number of uncompressed bytes input so far.
     *
     * <p>Since the number of bytes may be greater than
     * Integer.MAX_VALUE, the {@link #getBytesRead()} method is now
     * the preferred means of obtaining this information.</p>
     *
     * @return the total number of uncompressed bytes input so far
     */
    public int getTotalIn() {
	return (int) getBytesRead();
    }

    /**
     * Returns the total number of uncompressed bytes input so far.</p>
     *
     * @return the total (non-negative) number of uncompressed bytes input so far
     * @since 1.5
     */
    public long getBytesRead() {                                                //IBM-security_bulletin_tmp
        synchronized (zsRef){                                                   //IBM-security_bulletin_tmp
	    ensureOpen();                                                       //IBM-security_bulletin_tmp
	    return getBytesRead(zsRef.address());                               //IBM-security_bulletin_tmp
        }                                                                       //IBM-security_bulletin_tmp
    }

    /**
     * Returns the total number of compressed bytes output so far.
     *
     * <p>Since the number of bytes may be greater than
     * Integer.MAX_VALUE, the {@link #getBytesWritten()} method is now
     * the preferred means of obtaining this information.</p>
     *
     * @return the total number of compressed bytes output so far
     */
    public int getTotalOut() {
	return (int) getBytesWritten();
    }

    /**
     * Returns the total number of compressed bytes output so far.</p>
     *
     * @return the total (non-negative) number of compressed bytes output so far
     * @since 1.5
     */
    public long getBytesWritten() {                                             //IBM-security_bulletin_tmp
        synchronized (zsRef){                                                   //IBM-security_bulletin_tmp
	    ensureOpen();                                                       //IBM-security_bulletin_tmp
	    return getBytesWritten(zsRef.address());                            //IBM-security_bulletin_tmp
        }                                                                       //IBM-security_bulletin_tmp
    }

    /**
     * Resets deflater so that a new set of input data can be processed.
     * Keeps current compression level and strategy settings.
     */
    public void reset() {                                                       //IBM-security_bulletin_tmp
        synchronized (zsRef) {                                                  //IBM-security_bulletin_tmp
	    ensureOpen();                                                       //IBM-security_bulletin_tmp
	    reset(zsRef.address());                                             //IBM-security_bulletin_tmp
	    finish = false;                                                     //IBM-security_bulletin_tmp
	    finished = false;                                                   //IBM-security_bulletin_tmp
	    off = len = 0;                                                      //IBM-security_bulletin_tmp
        }                                                                       //IBM-security_bulletin_tmp
    }

    /**
     * Closes the compressor and discards any unprocessed input.
     * This method should be called when the compressor is no longer
     * being used, but will also be called automatically by the
     * finalize() method. Once this method is called, the behavior
     * of the Deflater object is undefined.
     */
    public void end() {                                                         //IBM-security_bulletin_tmp
        synchronized (zsRef) {                                                  //IBM-security_bulletin_tmp
            long addr = zsRef.address();                                        //IBM-security_bulletin_tmp
            zsRef.clear();                                                      //IBM-security_bulletin_tmp
	    if (addr != 0) {                                                    //IBM-security_bulletin_tmp
	        end(addr);                                                      //IBM-security_bulletin_tmp
	        buf = null;                                                     //IBM-security_bulletin_tmp
            }                                                                   //IBM-security_bulletin_tmp
	}
    }

    /**
     * Closes the compressor when garbage is collected.
     */
    protected void finalize() {
	end();
    }

    private void ensureOpen() {
        assert Thread.holdsLock(zsRef);                                         //IBM-security_bulletin_tmp
	if (zsRef.address() == 0)                                               //IBM-security_bulletin_tmp
	    throw new NullPointerException("Deflater has been closed");         //IBM-security_bulletin_tmp
    }

    private static native void initIDs();
    private native static long init(int level, int strategy, boolean nowrap);
    private native static void setDictionary(long addr, byte[] b, int off,      //IBM-security_bulletin_tmp
					     int len);
    private native int deflateBytes(long addr, byte[] b, int off, int len);     //IBM-security_bulletin_tmp
    private native static int getAdler(long addr);                              //IBM-security_bulletin_tmp
    private native static long getBytesRead(long addr);                         //IBM-security_bulletin_tmp
    private native static long getBytesWritten(long addr);                      //IBM-security_bulletin_tmp
    private native static void reset(long addr);                                //IBM-security_bulletin_tmp
    private native static void end(long addr);                                  //IBM-security_bulletin_tmp
}
//IBM-security_bulletin_tmp
