/* 
 *  Licensed to the Apache Software Foundation (ASF) under one or more 
 *  contributor license agreements.  See the NOTICE file distributed with 
 *  this work for additional information regarding copyright ownership. 
 *  The ASF licenses this file to You under the Apache License, Version 2.0 
 *  (the "License"); you may not use this file except in compliance with 
 *  the License.  You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License. 
 */

package javax.naming.ldap;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Enumeration;
import java.util.Iterator;

import javax.naming.InvalidNameException;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.BasicAttribute;
import javax.naming.directory.BasicAttributes;

import org.apache.harmony.jndi.internal.nls.Messages;
import org.apache.harmony.jndi.internal.parser.LdapRdnParser;
import org.apache.harmony.jndi.internal.parser.RdnAttribute;
import org.apache.harmony.jndi.internal.parser.RdnAttributeList;

/**
 * TODO: JavaDoc
 */
public class Rdn implements Serializable, Comparable<Object> {

    private static final long serialVersionUID = -5994465067210009656L;

    public static String escapeValue(Object val) {
        if (val == null) {
            throw new NullPointerException("val "
                    + Messages.getString("ldap.00"));
        }
        return LdapRdnParser.escapeValue(val);
    }

    public static Object unescapeValue(String val) {
        if (val == null) {
            throw new NullPointerException("val "
                    + Messages.getString("ldap.00"));
        }
        return LdapRdnParser.unescapeValue(val);
    }

    private transient RdnAttributeList list = null;
    
    private transient Attributes attributes = null;

    public Rdn(Attributes attrSet) throws InvalidNameException {
        if (attrSet == null) {
            throw new NullPointerException("attrSet "
                    + Messages.getString("ldap.00"));
        }

        if (attrSet.size() == 0) {
            throw new InvalidNameException("atrrSet "
                    + Messages.getString("ldap.03"));
        }

        // check all the elements to follow RI's behavior
        NamingEnumeration<? extends Attribute> ne = attrSet.getAll();
        while (ne.hasMoreElements()) {
            Attribute at = ne.nextElement();
            try {
                at.get();
            } catch (NamingException e) {
            }
        }

        attributes = (Attributes) attrSet.clone();
        toList();
    }

    public Rdn(Rdn rdn) {
        if (rdn == null) {
            throw new NullPointerException("rdn "
                    + Messages.getString("ldap.00"));
        }

        try {
            list = (RdnAttributeList) rdn.toList().clone() ;
        } catch (CloneNotSupportedException e) {
        }
    }

    public Rdn(String rdnString) throws InvalidNameException {
        if (rdnString == null) {
            throw new NullPointerException("rdnString "
                    + Messages.getString("ldap.00"));
        }

        if (rdnString.length() != 0) {
            String name = LdapRdnParser.parse(rdnString);
            list = LdapRdnParser.getListForRdn(name);
        } else {
            list = new RdnAttributeList();
        }
    }

    public Rdn(String type, Object value) throws InvalidNameException {
        if (type == null) {
            throw new NullPointerException("type "
                    + Messages.getString("ldap.00"));
        }

        if (value == null) {
            throw new NullPointerException("value "
                    + Messages.getString("ldap.00"));
        }

        if (type.length() == 0) {
            throw new InvalidNameException("type "
                    + Messages.getString("ldap.04"));
        }

        if (value instanceof String && ((String) value).length() == 0) {
            throw new InvalidNameException("value "
                    + Messages.getString("ldap.04"));
        }

        attributes = new BasicAttributes(type, value,
                true);
        toList();
    }

    public int compareTo(Object obj) {
        if (!(obj instanceof Rdn)) {
            throw new ClassCastException(Messages.getString("ldap.06")); //$NON-NLS-1$
        }
        Rdn rdn = (Rdn) obj;
        String s1 = ""; //$NON-NLS-1$
        String s2 = ""; //$NON-NLS-1$

        for (Enumeration<?> iter = attributes().getAll(); iter
                .hasMoreElements();) {
            s1 = s1 + escapeValue(iter.nextElement().toString());

            /*
             * this one does not seem necessary. Spec does not require it, if
             * there are apps that depend on commas, uncomment it
             */
            // if (iter.hasMoreElements()) {
            // s1 = s1 + ",";
            // }
        }
        for (Enumeration<?> iter = rdn.attributes().getAll(); iter
                .hasMoreElements();) {
            s2 = s2 + escapeValue(iter.nextElement().toString());

            /*
             * this one does not seem necessary. Spec does not require it, if
             * there are apps that depend on commas, uncomment it
             */
            // if (iter.hasMoreElements()) {
            // s2 = s2 + ",";
            // }
        }
        return s1.toLowerCase().compareTo(s2.toLowerCase());
    }

    private Attributes attributes() {
        if (attributes != null) {
            return (Attributes) attributes;
        }        
        return toAttributes();

    }

    public boolean equals(Object obj) {
        if (!(obj instanceof Rdn) || this.size() != ((Rdn) obj).size()) {
            return false;
        }

        if (this == obj) {
            return true;
        }

        int size = list.size();

        if (size != ((Rdn) obj).list.size()) {
            return false;
        }

        for (int i = 0; i < size; i++) {
            RdnAttribute ra = list.get(i);
            RdnAttribute otherRa = ((Rdn) obj).list.get(i);

            if (!ra.equals(otherRa)) {
                return false;
            }
        }

        return true;
    }

    public String getType() {
        return list.get(0).getID();
    }

    public Object getValue() {
        Object a = null;
        try {
            a = list.get(0).get();
        } catch (NullPointerException e) {
        }
        return a;
    }

    public int hashCode() {
        int sum = 0;

        for (Iterator<RdnAttribute> attr = list.iterator(); attr.hasNext();) {
            RdnAttribute a = attr.next();
            Object[] array = a.getAll();
            sum += a.getID().toLowerCase().hashCode();

            for (int i = 0; i < array.length; i++) {
                Object obj = array[i];
                if (obj instanceof byte[]) {
                    obj = new String((byte[]) obj);
                }
                try {
                    String s = (String) obj;
                    sum += escapeValue(s.toLowerCase()).hashCode();
                } catch (ClassCastException e) {
                    sum += obj.hashCode();
                }
            }
        }
        return sum;
    }

    public int size() {
        int result = 0;
        for (Iterator<RdnAttribute> iter = list.iterator(); iter.hasNext();) {
            result += iter.next().size();
        }
        return result;
    }

    public Attributes toAttributes() {
        if (attributes != null) {
            return (Attributes) attributes.clone();
        }
        BasicAttributes bas = new BasicAttributes(true);
        for (Iterator<RdnAttribute> iter = list.iterator(); iter.hasNext();) {
            RdnAttribute attr = iter.next();
            BasicAttribute ba = new BasicAttribute(attr.getID(), false);
            Object[] array = attr.getAll();
            
            for (int i = 0; i < array.length; i++) {
                ba.add(array[i]);
                
            }
            bas.put(ba);
        }
        return bas;
    }

    private RdnAttributeList toList() {
        if (list != null) {
            return list;
        }
        list = new RdnAttributeList();
        NamingEnumeration<? extends Attribute> ne = attributes.getAll();
        RdnAttribute ra = null;
        try {
            while (ne.hasMoreElements()) {
                Attribute attr = ne.nextElement();
                ra = new RdnAttribute(attr.getID());
                ra.add(attr.get());
                list.add(ra);
            }
        } catch (NamingException e) {

        }
        
        return list;
    }
    
    public String toString() {
        StringBuffer sb = new StringBuffer();
        for (Iterator<RdnAttribute> iter = list.iterator(); iter.hasNext();) {
            RdnAttribute element = iter.next();

            Object[] array = element.getAll();

            for (int i = 0; i < array.length; i++) {
                sb.append(element.getID());
                sb.append('=');
                sb.append(escapeValue(array[i]));

                if (i + 1 < array.length) {
                    sb.append('+');
                }
            }

            if (iter.hasNext()) {
                sb.append('+');
            }
        }
        return sb.toString();
    }

    private void readObject(ObjectInputStream ois) throws IOException,
            ClassNotFoundException, InvalidNameException {
        ois.defaultReadObject();
        String rdnString = (String) ois.readObject();
        if (rdnString == null) {
            throw new NullPointerException("rdnString "
                    + Messages.getString("ldap.00"));
        }
        if (rdnString.length() != 0) {
            String name = LdapRdnParser.parse(rdnString);
            list = LdapRdnParser.getListForRdn(name);
        } else {
            list = new RdnAttributeList();
        }
    }

    private void writeObject(ObjectOutputStream oos) throws IOException {
        oos.defaultWriteObject();
        oos.writeObject(this.toString());
    }

}
